import sys, struct, string
from types import *
from math import *
import os
from os import path
import md5
import gzip
import optimizer
from optimizer import *
import compress
from compress import *

def vertex_hash(position, normal, tangent, uv, extra_uv, color):
	return struct.pack("13i4B", int(position[0] * 1024), int(position[1] * 1024),
		int(position[2] * 1024), int(normal[0] * 1024), int(normal[1] * 1024),
		int(normal[2] * 1024), int(tangent[0] * 1024), int(tangent[1] * 1024),
		int(tangent[2] * 1024), int(uv[0] * 1024), int(uv[1] * 1024),
		int(extra_uv[0] * 1024), int(extra_uv[1] * 1024),
		int(color[0]), int(color[1]), int(color[2]), int(color[3]))

class el3d_error:
	msg = ""
	def __init__(self, msg_str):
		self.msg = msg_str
	def __repr__(self):
		return repr(self.msg)	

def build_options(has_normal, has_tangent, has_extra_uv, has_color, half_position, half_uv,
	half_extra_uv, compressed_normal, short_index):
	options = []
	if (has_normal != 0):
		options.append("has_normal")
	if (has_tangent != 0):
		options.append("has_tangent")
	if (has_extra_uv != 0):
		options.append("has_extra_uv")
	if (has_color != 0):
		options.append("has_color")
	if (half_position != 0):
		options.append("half_position")
	if (half_uv != 0):
		options.append("half_uv")
	if (half_extra_uv != 0):
		options.append("half_extra_uv")
	if (compressed_normal != 0):
		options.append("compressed_normal")
	if (short_index != 0):
		options.append("short_index")
	return options

def extract_options(options):
	has_normal = options.count("has_normal")
	has_tangent = options.count("has_tangent")
	has_extra_uv = options.count("has_extra_uv")
	has_color = options.count("has_color")
	half_position = options.count("half_position")
	half_uv = options.count("half_uv")
	half_extra_uv = options.count("half_extra_uv")
	compressed_normal = options.count("compressed_normal")
	short_index = options.count("short_index")
	return [has_normal, has_tangent, has_extra_uv, has_color, half_position, half_uv,
		half_extra_uv, compressed_normal, short_index]

class el3d_vertex:
	position = []
	normal = []
	uv = []
	tangent = []
	extra_uv = []
	color = []
	binary_format_position = "<3f"
	binary_format_normal = "<3f"
	binary_format_tangent = "<3f"
	binary_format_uv = "<2f"
	binary_format_extra_uv = "<2f"
	binary_format_color = "<4B"
	binary_format_half_position = "<3H"
	binary_format_compressed_normal = "<H"
	binary_format_compressed_tangent = "<H"
	binary_format_half_uv = "<2H"
	binary_format_half_extra_uv = "<2H"

	def __init__(self):
		self.position = [0] * 3
		self.normal = [0] * 3
		self.uv = [0] * 2
		self.tangent = [0] * 3
		self.extra_uv = [0] * 2
		self.color = [0] * 4
		self.normal[2] = 1.0
		self.tangent[0] = 1.0
		self.position = [0.0, 0.0, 0.0]
		self.normal = [0.0, 0.0, 1.0]
		self.uv = [0.0, 0.0]
		self.tangent = [1.0, 0.0, 0.0]
		self.extra_uv = [0.0, 0.0]
		self.color = [0, 0, 0, 0]

	def load(self, iofile, options, fix_uv):
		if (options.count("half_uv")):
			data = iofile.read(struct.calcsize(self.binary_format_half_uv))
			tmp = struct.unpack(self.binary_format_half_uv, data)
			self.uv[0] = vertex_compressor.half_to_float(tmp[0])
			self.uv[1] = vertex_compressor.half_to_float(tmp[1])
		else:
			data = iofile.read(struct.calcsize(self.binary_format_uv))
			tmp = struct.unpack(self.binary_format_uv, data)
			self.uv[0] = tmp[0]
			self.uv[1] = tmp[1]
		if (not fix_uv):
			self.uv[1] = 1.0 - self.uv[1]

		if (options.count("has_extra_uv")):
			if (options.count("half_extra_uv")):
				data = iofile.read(struct.calcsize(self.binary_format_half_extra_uv))
				tmp = struct.unpack(self.binary_format_half_extra_uv, data)
				self.extra_uv[0] = vertex_compressor.half_to_float(tmp[0])
				self.extra_uv[1] = vertex_compressor.half_to_float(tmp[1])
			else:
				data = iofile.read(struct.calcsize(self.binary_format_extra_uv))
				tmp = struct.unpack(self.binary_format_extra_uv, data)
				self.extra_uv[0] = tmp[0]
				self.extra_uv[1] = tmp[1]
			if (not fix_uv):
				self.extra_uv[1] = 1.0 - self.extra_uv[1]

		if (options.count("has_normal")):
			if (options.count("compressed_normal")):
				data = iofile.read(struct.calcsize(self.binary_format_compressed_normal))
				temp = struct.unpack(self.binary_format_compressed_normal, data)
				tmp = vertex_compressor.uncompress_normal(temp[0])
				self.normal[0] = tmp[0]
				self.normal[1] = tmp[1]
				self.normal[2] = tmp[2]
			else:
				data = iofile.read(struct.calcsize(self.binary_format_normal))
				tmp = struct.unpack(self.binary_format_normal, data)
				self.normal[0] = tmp[0]
				self.normal[1] = tmp[1]
				self.normal[2] = tmp[2]

		if (options.count("has_tangent")):
			if (options.count("compressed_normal")):
				data = iofile.read(struct.calcsize(self.binary_format_compressed_tangent))
				temp = struct.unpack(self.binary_format_compressed_tangent, data)
				tmp = vertex_compressor.uncompress_normal(temp[0])
				self.tangent[0] = tmp[0]
				self.tangent[1] = tmp[1]
				self.tangent[2] = tmp[2]
			else:
				data = iofile.read(struct.calcsize(self.binary_format_tangent))
				tmp = struct.unpack(self.binary_format_tangent, data)
				self.tangent[0] = tmp[0]
				self.tangent[1] = tmp[1]
				self.tangent[2] = tmp[2]

		if (options.count("half_position")):
			data = iofile.read(struct.calcsize(self.binary_format_half_position))
			tmp = struct.unpack(self.binary_format_half_position, data)
			self.position[0] = vertex_compressor.half_to_float(tmp[0])
			self.position[1] = vertex_compressor.half_to_float(tmp[1])
			self.position[2] = vertex_compressor.half_to_float(tmp[2])
		else:
			data = iofile.read(struct.calcsize(self.binary_format_position))
			tmp = struct.unpack(self.binary_format_position, data)
			self.position[0] = tmp[0]
			self.position[1] = tmp[1]
			self.position[2] = tmp[2]

		if (options.count("has_color")):
			data = iofile.read(struct.calcsize(self.binary_format_color))
			tmp = struct.unpack(self.binary_format_color, data)
			self.color[0] = tmp[0]
			self.color[1] = tmp[1]
			self.color[2] = tmp[2]
			self.color[3] = tmp[3]

	def save(self, iofile, options):
		tmp = [0] * 3
		tmp[0] = self.uv[0]
		tmp[1] = 1.0 - self.uv[1]

		if (options.count("half_uv")):
			tmp[0] = vertex_compressor.float_to_half(tmp[0])
			tmp[1] = vertex_compressor.float_to_half(tmp[1])
			data = struct.pack(self.binary_format_half_uv, tmp[0], tmp[1])
			iofile.write(data)
		else:
			data = struct.pack(self.binary_format_uv, tmp[0], tmp[1])
			iofile.write(data)
			
		if (options.count("has_extra_uv")):
			tmp[0] = self.extra_uv[0]
			tmp[1] = 1.0 - self.extra_uv[1]

			if (options.count("half_extra_uv")):
				tmp[0] = vertex_compressor.float_to_half(tmp[0])
				tmp[1] = vertex_compressor.float_to_half(tmp[1])
				data = struct.pack(self.binary_format_half_extra_uv, tmp[0], tmp[1])
				iofile.write(data)
			else:
				data = struct.pack(self.binary_format_extra_uv, tmp[0], tmp[1])
				iofile.write(data)

		if (options.count("has_normal")):
			if (options.count("compressed_normal")):
				tmp = vertex_compressor.compress_normal(self.normal)
				data = struct.pack(self.binary_format_compressed_normal, tmp)
				iofile.write(data)
			else:
				data = struct.pack(self.binary_format_normal, self.normal[0], self.normal[1], self.normal[2])
				iofile.write(data)

		if (options.count("has_tangent")):
			if (options.count("compressed_normal")):
				tmp = vertex_compressor.compress_normal(self.tangent)
				data = struct.pack(self.binary_format_compressed_tangent, tmp)
				iofile.write(data)
			else:
				data = struct.pack(self.binary_format_tangent, self.tangent[0], self.tangent[1], self.tangent[2])
				iofile.write(data)

		if (options.count("half_position")):
			tmp = [0] * 3
			tmp[0] = vertex_compressor.float_to_half(self.position[0])
			tmp[1] = vertex_compressor.float_to_half(self.position[1])
			tmp[2] = vertex_compressor.float_to_half(self.position[2])
			data = struct.pack(self.binary_format_half_position, tmp[0], tmp[1], tmp[2])
			iofile.write(data)
		else:
			data = struct.pack(self.binary_format_position, self.position[0], self.position[1], self.position[2])
			iofile.write(data)

		if (options.count("has_color")):
			data = struct.pack(self.binary_format_color, self.color[0], self.color[1], self.color[2], self.color[3])
			iofile.write(data)

	def size(self, options):
		ret = 0

		if (options.count("half_uv")):
			ret += struct.calcsize(self.binary_format_half_uv)
		else:
			ret += struct.calcsize(self.binary_format_uv)
			
		if (options.count("has_extra_uv")):
			if (options.count("half_extra_uv")):
				ret += struct.calcsize(self.binary_format_half_extra_uv)
			else:
				ret += struct.calcsize(self.binary_format_extra_uv)

		if (options.count("has_normal")):
			if (options.count("compressed_normal")):
				ret += struct.calcsize(self.binary_format_compressed_normal)
			else:
				ret += struct.calcsize(self.binary_format_normal)

		if (options.count("has_tangent")):
			if (options.count("compressed_normal")):
				ret += struct.calcsize(self.binary_format_compressed_tangent)
			else:
				ret += struct.calcsize(self.binary_format_tangent)

		if (options.count("half_position")):
			ret += struct.calcsize(self.binary_format_half_position)
		else:
			ret += struct.calcsize(self.binary_format_position)

		if (options.count("has_color")):
			ret += struct.calcsize(self.binary_format_color)

		return ret

class el3d_index:
	binary_format = "<1I"
	binary_format_short = "<1H"

	def load(self, iofile, options):
		if (options.count("short_index")):
			data = iofile.read(struct.calcsize(self.binary_format_short))
			return int(struct.unpack(self.binary_format_short, data)[0])
		else:
			data = iofile.read(struct.calcsize(self.binary_format))
			return int(struct.unpack(self.binary_format, data)[0])

	def save(self, iofile, index_val, options):
		if (options.count("short_index")):
			data = struct.pack(self.binary_format_short, index_val)
			iofile.write(data)
		else:
			data = struct.pack(self.binary_format, index_val)
			iofile.write(data)

	def size(self, options):
		if (options.count("short_index")):
			return struct.calcsize(self.binary_format_short)
		else:
			return struct.calcsize(self.binary_format)

class el3d_material:
	options = 0
	start = 0
	count = 0
	min_x = 0
	min_y = 0
	min_z = 0
	max_x = 0
	max_y = 0
	max_z = 0
	min_index = 0
	max_index = 0
	material_name = ""
	extra_material_name = ""
	binary_format_material = "<1i128s6f4i"
	binary_format_extra_material = "<128s"
	def load(self, iofile, options):
		data = iofile.read(struct.calcsize(self.binary_format_material))
		data = struct.unpack(self.binary_format_material, data)
		self.options = data[0]
		self.material_name = string.strip(data[1], '\0')
		self.min_x = data[2]
		self.min_y = data[3]
		self.min_z = data[4]
		self.max_x = data[5]
		self.max_y = data[6]
		self.max_z = data[7]
		self.min_index = data[8]
		self.max_index = data[9]
		self.start = data[10]
		self.count = data[11]
		if (options.count("has_extra_uv")):
			data = iofile.read(struct.calcsize(self.binary_format_extra_material))
			data = struct.unpack(self.binary_format_extra_material, data)
			self.extra_material_name = string.strip(data[0], '\0')
	def save(self, iofile, options):
		data = struct.pack(self.binary_format_material, self.options, \
			self.material_name, self.min_x, self.min_y, \
			self.min_z, self.max_x, self.max_y, self.max_z, \
			self.min_index, self.max_index, self.start, \
			self.count)
		iofile.write(data)
		if (options.count("has_extra_uv")):
			data = struct.pack(self.binary_format_extra_material, self.extra_material_name)
			iofile.write(data)
	def size(self, options):
		ret = struct.calcsize(self.binary_format_material)
		if (options.count("has_extra_uv")):
			ret += struct.calcsize(self.binary_format_extra_material)
		return ret

class el3d_header:
	magic = "e3dx"
	version_list = [[1, 0, 0, 0], [1, 0, 0, 1], [1, 1, 0, 0]]
	header_offset = 28
	index_no = 0
	index_size = 0
	index_offset = 0
	vertex_no = 0
	vertex_size = 0
	vertex_offset = 0
	material_no = 0
	material_size = 0
	material_offset = 0
	vertex_options = 0
	format_options = 0
	char_reserved_1 = 0
	char_reserved_2 = 0
	md5_obj = md5.new()
	binary_format = "<4c4B16B10i4B"
	def __init__(self):
		self.md5_obj = md5.new()
	def check_magic(self, d0, d1, d2, d3):
		if ((self.magic[0] == d0) and (self.magic[1] == d1) and
			(self.magic[2] == d2) and (self.magic[3] == d3)):
			return
		found_magic_string = str(d0) + str(d1) + str(d2) + str(d3)
		raise el3d_error("Error: File has wrong magic number! Found: \"" + \
			 found_magic_string + "\" Expected: \"" + self.magic + \
			"\"")
	def check_version(self, d0, d1, d2, d3, versions):
		for version in versions:
			if ((version[0] == d0) and (version[1] == d1) and
				(version[2] == d2) and (version[3] == d3)):
				return
		found_version = str(d0) + "." + str(d1) + "." + \
			str(d2) + "." + str(d3)
		expected_versions = []
		for i in range(0, len(versions)):
			expected_versions[i] = str(version[0]) + "." + \
				str(version[1]) + "." + str(version[2]) + \
				"." + str(version[3])
		expected_version = "["
		for i in range(0, len(versions) - 1):
			expected_version = expected_version + expected_versions[i] + ", "
		expected_version = expected_version + expected_versions[i] + "] "
		raise el3d_error("Error: File has wrong version number!" + \
			" Found: " + found_version + \
			" expected: " + expected_versions + "\"")
	def check_vertex_size(self, options):
		obj = el3d_vertex()
		tmp = obj.size(options)
		if (self.vertex_size != tmp):
			raise el3d_error("Error: File has wrong vertex size! Found: \"" + \
				str(self.vertex_size) + "\" Expected: \"" + str(tmp) \
				+ "\"")
	def check_index_size(self, options):
		obj = el3d_index()
		tmp = obj.size(options)
		if (self.index_size != tmp):
			raise el3d_error("Error: File has wrong index size! Found: \"" + \
				str(self.index_size) + "\" Expected: \"" + str(tmp) \
				+ "\"")
	def check_material_size(self, options):
		obj = el3d_material()
		tmp = obj.size(options)
		if (self.material_size != tmp):
			raise el3d_error("Error: File has wrong material size! Found: \"" + \
				str(self.material_size) + "\" Expected: \"" + str(tmp) \
				+ "\"")
	def get_digest(self):
		return struct.unpack("<16B", self.md5_obj.digest())

	def check_md5(self):
		tmp = self.get_digest()
		for i in range(0, 16):
			if (tmp[i] != self.digest[i]):
				raise el3d_error("Warning: File has wrong md5!")
	def load(self, iofile):
		data = iofile.read(self.size())
		data = struct.unpack(self.binary_format, data)
		self.check_magic(data[0], data[1], data[2], data[3])
		self.check_version(data[4], data[5], data[6], data[7], self.version_list)
		self.digest = [0] * 16
		self.digest[0] = data[8]
		self.digest[1] = data[9]
		self.digest[2] = data[10]
		self.digest[3] = data[11]
		self.digest[4] = data[12]
		self.digest[5] = data[13]
		self.digest[6] = data[14]
		self.digest[7] = data[15]
		self.digest[8] = data[16]
		self.digest[9] = data[17]
		self.digest[10] = data[18]
		self.digest[11] = data[19]
		self.digest[12] = data[20]
		self.digest[13] = data[21]
		self.digest[14] = data[22]
		self.digest[15] = data[23]
		self.header_offset = data[24]
		self.vertex_no = data[25]
		self.vertex_size = data[26]
		self.vertex_offset = data[27]
		self.index_no = data[28]
		self.index_size = data[29]
		self.index_offset = data[30]
		self.material_no = data[31]
		self.material_size = data[32]
		self.material_offset = data[33]
		self.vertex_options = data[34]
		self.format_options = data[35]
		self.char_reserved_1 = data[36]
		self.char_reserved_2 = data[37]

		if ((data[4] >= 1) & (data[5] >= 1)):
			has_normal = (self.vertex_options & 1) == 1
			fix_uv = 0
		else:
			has_normal = (self.vertex_options & 1) == 0
			self.format_options = 0
			fix_uv = 1

		has_tangent = (self.vertex_options / 2) & 1
		has_extra_uv = (self.vertex_options / 4) & 1
		has_color = (self.vertex_options / 8) & 1

		half_position = (self.format_options / 1) & 1
		half_uv = (self.format_options / 2) & 1
		half_extra_uv = (self.format_options / 4) & 1
		compressed_normal = (self.format_options / 8) & 1
		short_index = (self.format_options / 16) & 1

		options = build_options(has_normal, has_tangent, has_extra_uv, has_color, half_position,
			half_uv, half_extra_uv, compressed_normal, short_index)

		self.check_vertex_size(options)
		self.check_index_size(options)
		self.check_material_size(options)

		return options, fix_uv
	def save(self, iofile, version_index):
		self.header_offset = 28
		digest = self.get_digest()
		version = self.version_list[version_index]
		data = struct.pack(self.binary_format, self.magic[0], \
			self.magic[1], self.magic[2], self.magic[3], \
			version[0], version[1], version[2], version[3],	\
			digest[0], digest[1], digest[2], digest[3],	\
			digest[4], digest[5], digest[6], digest[7],	\
			digest[8], digest[9], digest[10], digest[11],	\
			digest[12], digest[13], digest[14], digest[15],	\
			self.header_offset, self.vertex_no, \
			self.vertex_size, self.vertex_offset, self.index_no, \
			self.index_size, self.index_offset, self.material_no, \
			self.material_size, self.material_offset, \
			self.vertex_options, self.format_options, \
			self.char_reserved_1, self.char_reserved_2)
		iofile.write(data)
	def size(self):
		return struct.calcsize(self.binary_format)
	def build_md5(self, iofile):
		self.md5_obj = md5.new()
		iofile.seek(self.header_offset)
		data = iofile.read()
		self.md5_obj.update(data)

class el3d_face:
	index = []
	def __init__(self):
		self.index = [0] * 3

class el3d_file:
	options = []
	vertex_list = []
	index_list = []
	face_list = []
	material_list = []
	vertex_dict = {}
	data_list = []
	data_dict = {}
	header = el3d_header()
	def __init__(self):
		self.options = []
		self.vertex_list = []
		self.index_list = []
		self.face_list = []
		self.material_list = []
		self.vertex_dict = {}
		self.data_list = []
		self.data_dict = {}
		self.header = el3d_header()
	def load(self, iofile):
		self.vertex_list = []
		self.index_list = []
		self.material_list = []
		self.header = el3d_header()
		self.options, fix_uv = self.header.load(iofile)
		self.header.build_md5(iofile)
		self.header.check_md5()
		iofile.seek(self.header.vertex_offset)
		#read vertices
		for i in range(0, self.header.vertex_no):
			self.vertex_list.append(el3d_vertex())
			self.vertex_list[i].load(iofile, self.options, fix_uv)
		iofile.seek(self.header.index_offset)
		#read indicies
		el3d_idx = el3d_index()
		for i in range(0, self.header.index_no):
			self.index_list.append(el3d_idx.load(iofile, self.options))
		iofile.seek(self.header.material_offset)
		for i in range(0, self.header.material_no):
			self.material_list.append(el3d_material())
			self.material_list[i].load(iofile, self.options)
		iofile.close()

	def optimize(self):
		for material in self.material_list:
			idx_list = [0] * material.count
			for i in range(0, material.count):
				idx_list[i] = self.index_list[material.start + i]
			optimize_vertex_cache_order(idx_list, 128)
			for i in range(0, material.count):
				self.index_list[material.start + i] = idx_list[i]

	def check_half_position(self):
		ret = True
		for vertex in self.vertex_list:
			ret &= abs(vertex.position[0]) <= 65504.0
			ret &= abs(vertex.position[1]) <= 65504.0
			ret &= abs(vertex.position[2]) <= 65504.0

		return ret

	def check_half_uv(self):
		ret = True
		for vertex in self.vertex_list:
			ret &= abs(vertex.uv[0]) <= 65504.0
			ret &= abs(vertex.uv[1]) <= 65504.0

		return ret

	def check_half_extra_uv(self):
		ret = True
		for vertex in self.vertex_list:
			ret &= abs(vertex.extra_uv[0]) <= 65504.0
			ret &= abs(vertex.extra_uv[1]) <= 65504.0

		return ret

	def save(self, filename, has_extra_uv, has_color):
		self.header = el3d_header()
		self.optimize()

		half_position = self.check_half_position()
		half_uv = self.check_half_uv()
		half_extra_uv = self.check_half_extra_uv()
		short_index = len(self.vertex_list) <= 65536

		options = build_options(1, 1, has_extra_uv, has_color, half_position,
			half_uv, half_extra_uv, 1, short_index)

		self.header.vertex_options = 0
		self.header.format_options = 0
		extracted_options = extract_options(options)
		if (extracted_options[0]):
			self.header.vertex_options += 1
		if (extracted_options[1]):
			self.header.vertex_options += 2
		if (extracted_options[2]):
			self.header.vertex_options += 4
		if (extracted_options[3]):
			self.header.vertex_options += 8
		if (extracted_options[4]):
			self.header.format_options += 1
		if (extracted_options[5]):
			self.header.format_options += 2
		if (extracted_options[6]):
			self.header.format_options += 4
		if (extracted_options[7]):
			self.header.format_options += 8
		if (extracted_options[8]):
			self.header.format_options += 16

		print "vertexes: ", len(self.vertex_list)
		print "indexes: ", len(self.index_list)
		print "materials: ", len(self.material_list)
		self.header.index_no = len(self.index_list)
		self.header.vertex_no = len(self.vertex_list)
		self.header.material_no = len(self.material_list)
		tmp = el3d_vertex()
		self.header.vertex_size = tmp.size(options)
		tmp = el3d_material()
		self.header.material_size = tmp.size(options)
		el3d_idx = el3d_index()
		self.header.index_size = el3d_idx.size(options)
		self.header.vertex_offset = self.header.size()
		self.header.index_offset = self.header.vertex_offset + self.header.vertex_no * self.header.vertex_size
		self.header.material_offset = self.header.index_offset + self.header.index_no * self.header.index_size

		iofile = os.tmpfile()
		self.header.save(iofile, 2)
		for v in self.vertex_list:
			v.save(iofile, options)
		for i in self.index_list:
			el3d_idx.save(iofile, i, options)
		for m in self.material_list:
			m.save(iofile, options)
		self.header.build_md5(iofile)
		iofile.close()

		iofile = open(filename, "wb")
		self.header.save(iofile, 2)
		for v in self.vertex_list:
			v.save(iofile, options)
		for i in self.index_list:
			el3d_idx.save(iofile, i, options)
		for m in self.material_list:
			m.save(iofile, options)
		iofile.close()

	def build_vertex_md5(self):
		md5_obj = md5.new()
		for hash_value in self.vertex_dict:
			md5_obj.update(hash_value)
		return md5_obj.digest()
	def add_vertex(self, position, normal, tangent, uv, extra_uv, color):
		hash_value = vertex_hash(position, normal, tangent, uv, extra_uv, color)
		if not self.vertex_dict.has_key(hash_value):
			index = len(self.vertex_list)
			self.vertex_dict[hash_value] = index
			self.vertex_list.append(el3d_vertex())
			self.vertex_list[index].position = position
			self.vertex_list[index].normal = normal
			self.vertex_list[index].tangent = tangent
			self.vertex_list[index].uv = uv
			self.vertex_list[index].extra_uv = extra_uv
			self.vertex_list[index].color = color
			print "adding vertex: ", len(self.vertex_list)
		index = self.vertex_dict[hash_value]
		return index
	def add_index(self, index):
		self.index_list.append(index)
	def add_face_corner(self, position, normal, tangent, uv, extra_uv, color):
		index = self.add_vertex(position, normal, tangent, uv, extra_uv, color)
		self.add_index(index)
		idx = len(self.material_list) - 1
		self.material_list[idx].min_index = min(self.material_list[idx].min_index, index)
		self.material_list[idx].max_index = max(self.material_list[idx].max_index, index)
		if (self.vertex_list[index].position[0] < self.material_list[idx].min_x):
			self.material_list[idx].min_x = self.vertex_list[index].position[0]
		if (self.vertex_list[index].position[1] < self.material_list[idx].min_y):
			self.material_list[idx].min_y = self.vertex_list[index].position[1]
		if (self.vertex_list[index].position[2] < self.material_list[idx].min_z):
			self.material_list[idx].min_z = self.vertex_list[index].position[2]
		if (self.vertex_list[index].position[0] > self.material_list[idx].max_x):
			self.material_list[idx].max_x = self.vertex_list[index].position[0]
		if (self.vertex_list[index].position[1] > self.material_list[idx].max_y):
			self.material_list[idx].max_y = self.vertex_list[index].position[1]
		if (self.vertex_list[index].position[2] > self.material_list[idx].max_z):
			self.material_list[idx].max_z = self.vertex_list[index].position[2]
		self.material_list[idx].count += 1
	def add_material(self, material_name, extra_material_name, options):
		idx = len(self.material_list)
		self.material_list.append(el3d_material())
		self.material_list[idx].material_name = material_name
		self.material_list[idx].extra_material_name = extra_material_name
		self.material_list[idx].options = options
		self.material_list[idx].start = len(self.index_list)
		self.material_list[idx].count = 0
		self.material_list[idx].min_index = 0xFFFFFFFF
		self.material_list[idx].max_index = 0
		self.material_list[idx].min_x = 10e30
 		self.material_list[idx].min_y = 10e30
		self.material_list[idx].min_z = 10e30
		self.material_list[idx].max_x = -10e30
 		self.material_list[idx].max_y = -10e30
		self.material_list[idx].max_z = -10e30

def zeros(n):
	return vector(map(lambda x: 0.0, range(n)))

class vector(list):
	def __getitem__(self, index):
		return self.data[index]
	def __setitem__(self, index, value):
		self.data[index] = value
	def __len__(self):
		return len(self.data)
	def __init__(self, data):
		self.data = data
	def __repr__(self):
		return repr(self.data)
	def __add__(self, other):
		return vector(map(lambda x, y: x + y, self.data, other.data))
	def __sub__(self, other):
		return vector(map(lambda x, y: x - y, self.data, other.data))
	def __mul__(self, other):
		try:
			return vector(map(lambda x, y: x * y, self.data, other.data))
		except:
			return vector(map(lambda x: x * other, self.data))
	def __rmul__(self, other):
		return (self * other)
	def __div__(self, other):
		try:
			return vector(map(lambda x, y: x / y, self.data, other.data))
		except:
			return vector(map(lambda x: x / other, self.data))
	def __rdiv__(self, other):
	    try:
		    return vector(map(lambda x, y: x / y, other.data, self.data))
	    except:
		    return vector(map(lambda x: other / x, self.data))
	def __neg__(self):
		return vector(map(lambda x: -x, self.data))
	def dot(self, other):
		try:
			return reduce(lambda x, y: x + y, (self * other).data)
		except:
			raise TypeError, 'vector::FAILURE in dot'
	def vlen(self):
		return sqrt(self.dot(self))
	def normalize(self):
		scale = self.vlen()
		if (scale > 1e-7):
			self.data = map(lambda x: x / scale, self.data)
		else:
			self[0] = 0.0
			self[1] = 0.0
			self[2] = 1.0
	def cross(self, other):
		return vector([self[1] * other[2] - self[2] * other[1],
				self[2] * other[0] - self[0] * other[2],
				self[0] * other[1] - self[1] * other[0]])
	def size(self):
		return len(self.data)

